# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2023 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Classes to encapsulate the Addon Manager's interaction with FreeCAD, and to provide
replacements when the Addon Manager is not run from within FreeCAD (e.g. during unit
testing).

Usage:
from addonmanager_freecad_interface import Console, DataPaths, Preferences
"""

import json
import logging
import os
import platform
import shutil
import tempfile

# pylint: disable=too-few-public-methods

try:
    import FreeCAD

    try:
        from freecad.utils import get_python_exe

        try:
            _ = get_python_exe()
        except AttributeError as e:
            raise RuntimeError("Could not get the FreeCAD python executable") from e
    except ImportError:
        # This was only added in FreeCAD 1.0 -- to support FreeCAD 0.21 a backup strategy must be
        # used. This code is borrowed from FreeCAD 1.1dev.
        def get_python_exe():
            prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/PythonConsole")
            python_exe = prefs.GetString("ExternalPythonExecutable", "Not set")
            fc_dir = FreeCAD.getHomePath()
            if not python_exe or python_exe == "Not set" or not os.path.exists(python_exe):
                python_exe = os.path.join(fc_dir, "bin", "python3")
                if "Windows" in platform.system():
                    python_exe += ".exe"

            if not python_exe or not os.path.exists(python_exe):
                python_exe = os.path.join(fc_dir, "bin", "python")
                if "Windows" in platform.system():
                    python_exe += ".exe"

            if not python_exe or not os.path.exists(python_exe):
                python_exe = shutil.which("python3")

            if not python_exe or not os.path.exists(python_exe):
                python_exe = shutil.which("python")

            if not python_exe or not os.path.exists(python_exe):
                return ""

            python_exe = python_exe.replace("/", os.path.sep)
            prefs.SetString("ExternalPythonExecutable", python_exe)
            return python_exe

    if not hasattr(FreeCAD, "Console"):
        raise ImportError("Unrecognized FreeCAD version")

    Console = FreeCAD.Console
    ParamGet = FreeCAD.ParamGet
    Version = FreeCAD.Version
    getUserAppDataDir = FreeCAD.getUserAppDataDir
    getUserMacroDir = FreeCAD.getUserMacroDir
    getUserCachePath = FreeCAD.getUserCachePath
    translate = FreeCAD.Qt.translate
    loadUi = None
    GuiUp = FreeCAD.GuiUp

    if GuiUp:
        import FreeCADGui

        if hasattr(FreeCADGui, "PySideUic"):
            loadUi = FreeCADGui.PySideUic.loadUi

        addPreferencePage = FreeCADGui.addPreferencePage
    else:
        FreeCADGui = None

except ImportError:
    FreeCAD = None
    FreeCADGui = None
    getUserAppDataDir = None
    getUserCachePath = None
    getUserMacroDir = None
    loadUi = None

    try:
        from PySide6 import QtCore, QtWidgets

        GuiUp = (
            True
            if hasattr(QtWidgets, "QApplication") and QtWidgets.QApplication.instance()
            else False
        )
        from PySide6.QtUiTools import QUiLoader
    except ImportError:
        try:
            from PySide2 import QtCore, QtWidgets

            GuiUp = True if QtWidgets.QApplication.instance() else False
            from PySide2.QtUiTools import QUiLoader
        except ImportError:
            GuiUp = False

    if GuiUp:

        def loadUi(path: str):
            loader = QUiLoader()
            file = QtCore.QFile(path)
            file.open(QtCore.QFile.ReadOnly)
            window = loader.load(file)
            file.close()
            return window

        def addPreferencePage(_options_class, _name: str):
            """Don't do anything with a preference page right now"""
            pass

    def translate(context: str, sourceText: str, disambiguation: str = "", n: int = -1) -> str:
        return QtCore.QCoreApplication.translate(context, sourceText, disambiguation, n)

    def Version():
        return 1, 1, 0, "dev"

    def get_python_exe():
        return shutil.which("python3")

    class ConsoleReplacement:
        """If FreeCAD's Console is not available, create a replacement by redirecting FreeCAD
        log calls to Python's built-in logging facility."""

        @staticmethod
        def PrintLog(arg: str) -> None:
            logging.log(logging.DEBUG, arg)

        @staticmethod
        def PrintMessage(arg: str) -> None:
            logging.info(arg)

        @staticmethod
        def PrintWarning(arg: str) -> None:
            logging.warning(arg)

        @staticmethod
        def PrintError(arg: str) -> None:
            logging.error(arg)

    Console = ConsoleReplacement()

    class ParametersReplacement:
        """Proxy for FreeCAD's Parameters when not running within FreeCAD. NOT
        serialized, only exists for the duration of the program's execution. Only
        provides the functions used by the Addon Manager, this class is not intended
        to be a complete replacement for FreeCAD's preferences system."""

        parameters = {}

        def GetBool(self, name: str, default: bool) -> bool:
            return self._Get(name, default)

        def GetInt(self, name: str, default: int) -> int:
            return self._Get(name, default)

        def GetFloat(self, name: str, default: float) -> float:
            return self._Get(name, default)

        def GetString(self, name: str, default: str) -> str:
            return self._Get(name, default)

        def _Get(self, name, default):
            return self.parameters[name] if name in self.parameters else default

        def SetBool(self, name: str, value: bool) -> None:
            self.parameters[name] = value

        def SetInt(self, name: str, value: int) -> None:
            self.parameters[name] = value

        def SetFloat(self, name: str, value: float) -> None:
            self.parameters[name] = value

        def SetString(self, name: str, value: str) -> None:
            self.parameters[name] = value

        def RemBool(self, name: str) -> None:
            self.parameters.pop(name)

        def RemInt(self, name: str) -> None:
            self.parameters.pop(name)

        def RemFloat(self, name: str) -> None:
            self.parameters.pop(name)

        def RemString(self, name: str) -> None:
            self.parameters.pop(name)

    def ParamGet(_: str):
        return ParametersReplacement()


class DataPaths:
    """Provide access to various data storage paths. If not running within FreeCAD,
    all paths are temp directories. If not run within FreeCAD, all directories are
    deleted when the last reference to this class is deleted."""

    data_dir = None
    mod_dir = None
    macro_dir = None
    cache_dir = None
    home_dir = None

    reference_count = 0

    def __init__(self):
        if FreeCAD:
            if self.data_dir is None:
                self.data_dir = getUserAppDataDir()
            if self.mod_dir is None:
                self.mod_dir = os.path.join(getUserAppDataDir(), "Mod")
            if self.cache_dir is None:
                # Anytime the cache format changes, increment this version number so we don't
                # interfere with old versions.
                self.cache_dir = os.path.join(getUserCachePath(), "AddonManager2026-1")
            if self.macro_dir is None:
                self.macro_dir = getUserMacroDir(True)
            if self.home_dir is None:
                self.home_dir = FreeCAD.getHomePath()
        else:
            self.reference_count += 1
            if self.data_dir is None:
                self.data_dir = tempfile.mkdtemp()
            if self.mod_dir is None:
                self.mod_dir = tempfile.mkdtemp()
            if self.cache_dir is None:
                self.cache_dir = tempfile.mkdtemp()
            if self.macro_dir is None:
                self.macro_dir = tempfile.mkdtemp()
            if self.home_dir is None:
                self.home_dir = os.path.join(os.path.dirname(__file__))

    def __del__(self):
        self.reference_count -= 1
        if not FreeCAD and self.reference_count <= 0:
            self._delete_paths()

    def _delete_paths(self):
        if FreeCAD:
            return
        paths = [self.data_dir, self.mod_dir, self.cache_dir, self.macro_dir, self.mod_dir]
        for path in paths:
            if os.path.isdir(path):
                os.rmdir(path)
        self.data_dir = None
        self.mod_dir = None
        self.cache_dir = None
        self.macro_dir = None


class Preferences:
    """Wrap access to all user preferences. If run within FreeCAD, user preferences are
    persistent, otherwise they only exist per-run. All preferences are controlled by a
    central JSON file defining their defaults."""

    preferences_defaults = {}

    def __init__(self, defaults_data=None):
        """Set up the preferences, initializing the class statics if necessary. If
        defaults_data is provided it is used as the preferences defaults. If it is not
        provided, then the defaults are read in from the standard defaults file,
        addonmanager_preferences_defaults.json, located in the same directory as this
        Python file."""
        if not self.preferences_defaults:
            if defaults_data:
                self.preferences_defaults = defaults_data
            else:
                self._load_preferences_defaults()
        self.prefs = ParamGet("User parameter:BaseApp/Preferences/Addons")

    def get(self, name: str):
        """Get the preference value for the given key"""
        if name not in self.preferences_defaults:
            raise RuntimeError(
                f"Unrecognized preference {name} -- did you add "
                + "it to addonmanager_preferences_defaults.json?"
            )
        if isinstance(self.preferences_defaults[name], bool):
            return self.prefs.GetBool(name, self.preferences_defaults[name])
        if isinstance(self.preferences_defaults[name], int):
            return self.prefs.GetInt(name, self.preferences_defaults[name])
        if isinstance(self.preferences_defaults[name], float):
            return self.prefs.GetFloat(name, self.preferences_defaults[name])
        if isinstance(self.preferences_defaults[name], str):
            return self.prefs.GetString(name, self.preferences_defaults[name])
        # We don't directly support any other types from the JSON file (e.g. arrays)
        type_name = type(self.preferences_defaults[name])
        raise RuntimeError(f"Unrecognized type for {name}: {type_name}")

    def set(self, name: str, value):
        """Set the preference value for the given key. Must exist (e.g. must be in the
        addonmanager_preferences_defaults.json file)."""
        if name not in self.preferences_defaults:
            raise RuntimeError(
                f"Unrecognized preference {name} -- did you add "
                + "it to addonmanager_preferences_defaults.json?"
            )
        if isinstance(self.preferences_defaults[name], bool):
            self.prefs.SetBool(name, value)
        elif isinstance(self.preferences_defaults[name], int):
            self.prefs.SetInt(name, value)
        elif isinstance(self.preferences_defaults[name], float):
            self.prefs.SetFloat(name, value)
        elif isinstance(self.preferences_defaults[name], str):
            self.prefs.SetString(name, value)
        else:
            # We don't directly support any other types from the JSON file (e.g. arrays)
            type_name = type(self.preferences_defaults[name])
            raise RuntimeError(f"Unrecognized type for {name}: {type_name}")

    def rem(self, name: str):
        """Remove the preference. Must have an entry in the
        addonmanager_preferences_defaults.json file."""
        if name not in self.preferences_defaults:
            raise RuntimeError(
                f"Unrecognized preference {name} -- did you add "
                + "it to addonmanager_preferences_defaults.json?"
            )
        if isinstance(self.preferences_defaults[name], bool):
            return self.prefs.RemBool(name)
        if isinstance(self.preferences_defaults[name], int):
            return self.prefs.RemInt(name)
        if isinstance(self.preferences_defaults[name], float):
            return self.prefs.RemFloat(name)
        if isinstance(self.preferences_defaults[name], str):
            return self.prefs.RemString(name)
        # We don't directly support any other types from the JSON file (e.g. arrays)
        type_name = type(self.preferences_defaults[name])
        raise RuntimeError(f"Unrecognized type for {name}: {type_name}")

    @classmethod
    def _load_preferences_defaults(cls, filename=None):
        """Loads the preferences defaults JSON file from either a specified file, or
        from the standard addonmanager_preferences_defaults.json file."""

        if filename is None:
            json_file = os.path.join(
                os.path.dirname(__file__), "addonmanager_preferences_defaults.json"
            )
        else:
            json_file = filename
        with open(json_file, "r", encoding="utf-8") as f:
            file_contents = f.read()
        cls.preferences_defaults = json.loads(file_contents)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Jon Nordby <jononor@gmail.com>                     *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


TOOLTIP = """
FreeCAD Path post-processor to output code for the Roland Modela MDX-## machines.

The machine speaks RML-1, specified in 'Roland RML-1 Programming Guidelines'
https://itp.nyu.edu/classes/cdp-spring2014/files/2014/09/RML1_Manual.pdf
http://altlab.org/d/content/m/pangelo/ideas/rml_command_guide_en_v100.pdf

The format has some overlap with HPGL:
https://en.wikipedia.org/wiki/HPGL
http://paulbourke.net/dataformats/hpgl/
"""

import FreeCAD
import Part
import Path.Post.Utils as PostUtils
from builtins import open as pyopen


# Entrypoint used by FreeCAD
def export(objectslist, filename, argstring):
    "Export objects as Roland Modela code."

    code = ""
    for obj in objectslist:
        code += convertobject(obj)

    gfile = pyopen(filename, "w")
    gfile.write(code)
    gfile.close()


def convertobject(obj):
    gcode = obj.Path.toGCode()
    gcode = parse(gcode)
    return gcode


def motoron():
    return ["!MC1;"]


def motoroff():
    return ["!MC0;"]


def home():
    return ["H;"]


def setjog():
    # "!PZ%d,%d;",iz_down,iz_up); // set z down, jog
    return ""


def addheader():
    return ["PA;PA;"]  # absolute positioning


def addfooter():
    return []


def mm2cord(mm):
    mm = float(mm)
    return int(40.0 * mm)


def feed(x=None, y=None, z=None, state=None):
    c = []
    if state is None:
        state = {}

    if x is not None:
        x = float(x)
        state["X"] = x
    if y is not None:
        y = float(y)
        state["Y"] = y
    if z is not None:
        z = float(z)
        state["Z"] = z

    if x is not None and y is not None and z is not None:
        # 3d motion
        c.append("Z%d,%d,%d;" % (mm2cord(x), mm2cord(y), mm2cord(z)))
    elif x is not None and y is not None:
        # 2d in XY plane
        c.append("PD%d,%d;" % (mm2cord(x), mm2cord(y)))
    elif z is not None:
        pass
    return c


def jog(x=None, y=None, z=None, state=None):
    c = []
    if state is None:
        state = {}
    if x is not None and y is not None:
        x, y = float(x), float(y)
        c.append("PU%d,%d;" % (mm2cord(x), mm2cord(y)))
        state["X"] = x
        state["Y"] = y
    if z is not None:
        z = float(z)
        c.append("PU;")
        state["Z"] = z

    return c


def xyarc(args, state):
    # no native support in RML/Modela, convert to linear line segments
    c = []

    lastPoint = FreeCAD.Vector(state["X"], state["Y"])
    newPoint = FreeCAD.Vector(float(args["X"]), float(args["Y"]))
    centerOffset = FreeCAD.Vector(float(args["I"]), float(args["J"]))
    center = lastPoint + centerOffset
    radius = (center - lastPoint).Length
    xyNormal = FreeCAD.Vector(0, 0, 1)
    circle = Part.Circle(center, xyNormal, radius)
    p0 = circle.parameter(lastPoint)
    p1 = circle.parameter(newPoint)
    arc = Part.ArcOfCircle(circle, p0, p1)
    steps = 64  # specify max error instead?
    points = arc.discretize(steps)
    # consider direction?
    # print('p = Part.ArcOfCircle(Part.Circle(FreeCAD.Vector(%f, %f), FreeCAD.Vector(0, 0, 1), %f), %f, %f)' % (center.x, center.y, radius, p0, p1))
    for p in points:
        c += feed(p.x, p.y, state["Z"], state)
    return c


def speed(xy=None, z=None, state=None):
    c = []
    if state is None:
        state = {}
    print(xy, z, state)
    if xy is not None:
        xy = float(xy)
        if xy > 0.0 and xy != state["XYspeed"]:
            c.append("VS%.1f;" % xy)
            state["XYspeed"] = xy
    if z is not None:
        z = float(z)
        if z > 0.0 and z != state["Zspeed"]:
            c.append("!VZ%.1f;" % z)
            state["Zspeed"] = z
    return c


def convertgcode(cmd, args, state):
    """Convert a single gcode command to equivalent Roland code"""
    if cmd == "G0":
        # jog
        return jog(args["X"], args["Y"], args["Z"], state)
    elif cmd == "G1":
        # linear feed
        c = []
        # feedrate
        c += speed(xy=args["F"], z=args["F"], state=state)
        # motion
        c += feed(args["X"], args["Y"], args["Z"], state)
        return c
    elif cmd == "G2" or cmd == "G3":
        # arc feed
        c = []
        # feedrate
        c += speed(xy=args["F"], state=state)
        # motion
        if args["X"] and args["Y"] and args["Z"]:
            # helical motion
            pass
        elif args["X"] and args["Y"]:
            # arc in plane
            c += xyarc(args, state)
        return c
    elif cmd == "G20":
        # inches mode
        raise ValueError("rml_post: Inches mode not supported")
    elif cmd == "G21":
        # millimeter mode
        return ""
    elif cmd == "G40":
        # tool compensation off
        return ""
    elif cmd == "G80":
        # cancel all cycles (drill normally)
        return "PU;"
    elif cmd == "G81":
        c = []
        # feedrate
        c += speed(z=args["F"], state=state)
        # motion
        c += jog(args["X"], args["Y"], state=state)
        c += feed(args["X"], args["Y"], args["Z"], state)
        return c
    elif cmd == "G90":
        # absolute mode?
        return ""
    elif cmd == "G98":
        # feedrate
        return ""
    else:
        raise NotImplementedError("rml_post: GCode command %s not understood" % (cmd,))


def parse(inputstring):
    "parse(inputstring): returns a parsed output string"

    state = {"X": 0.0, "Y": 0.0, "Z": 0.0, "XYspeed": -1.0, "Zspeed": -1.0}
    output = []

    # header
    output += addheader()
    output += motoron()

    output += speed(2.0, 1.0, state)  # defaults

    # respect clearance height?

    # treat the input line by line
    lines = inputstring.split("\n")
    for line in lines:
        if not line:
            continue
        parsed = PostUtils.stringsplit(line)
        command = parsed["command"]
        print("cmd", line)
        try:
            if command:
                code = convertgcode(command, parsed, state)
                if not isinstance(code, list):
                    code = [code]
                if len(code) and code[0]:
                    output += code
        except NotImplementedError as e:
            print(e)

    # footer
    output += motoroff()
    output += home()
    output += addfooter()

    return "\n".join(output)


# print (__name__ + " gcode postprocessor loaded.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Unit tests for the Draft Workbench, Coin (Pivy) tests."""

## @package test_pivy
# \ingroup drafttests
# \brief Unit tests for the Draft Workbench, Coin (Pivy) tests.

## \addtogroup drafttests
# @{

import FreeCADGui as Gui
from drafttests import auxiliary as aux
from drafttests import test_base
from draftutils.messages import _msg


class DraftPivy(test_base.DraftTestCaseDoc):
    """Test for the presence of Pivy and that it works with Coin3D."""

    def test_pivy_import(self):
        """Import Coin (Pivy)."""
        module = "pivy.coin"
        imported = aux.import_test(module)
        self.assertTrue(imported, "Problem importing '{}'".format(module))

    def test_pivy_draw(self):
        """Use Coin (pivy.coin) to draw a cube on the active view."""
        import pivy.coin as coin

        cube = coin.SoCube()
        _msg("  Draw cube")
        Gui.getDocument(self.doc).ActiveView.getSceneGraph().addChild(cube)
        _msg("  Adding cube to the active view scene")
        self.assertTrue(cube, "Pivy is not working properly.")


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with cubic shapes (parallelepipeds)."""
## @package cuboids
# \ingroup draftgeoutils
# \brief Provides various functions for cubic shapes (parallelepipeds).

## \addtogroup draftgeoutils
# @{
import math

import FreeCAD as App
import DraftVecUtils

from draftgeoutils.general import geomType, vec, precision


def isCubic(shape):
    """Return True if the shape is a parallelepiped (cuboid).

    A parallelepiped of cube-like shape has 8 vertices, 6 faces, 12 edges,
    and all angles are 90 degrees between its edges.
    """
    # first we try fast methods
    if len(shape.Vertexes) != 8 or len(shape.Faces) != 6 or len(shape.Edges) != 12:
        return False

    for e in shape.Edges:
        if geomType(e) != "Line":
            return False

    # if ok until now, let's do more advanced testing
    for f in shape.Faces:
        if len(f.Edges) != 4:
            return False

        for i in range(4):
            e1 = vec(f.Edges[i])
            if i < 3:
                e2 = vec(f.Edges[i + 1])
            else:
                e2 = vec(f.Edges[0])
            rpi = [0.0, round(math.pi / 2, precision())]
            if round(e1.getAngle(e2), precision()) not in rpi:
                return False

    return True


def getCubicDimensions(shape):
    """Return a list containing the placement, and dimensions of the shape.

    The dimensios are length, width and height of a the parallelepiped,
    rounded to the value indicated by `precision`.
    The placement point is the lowest corner of the shape.

    If it is not a parallelepiped (cuboid), return None.
    """
    if not isCubic(shape):
        return None

    # determine lowest face, which will be our base
    z = [10, 1000000000000]
    for i in range(len(shape.Faces)):
        if shape.Faces[i].CenterOfMass.z < z[1]:
            z = [i, shape.Faces[i].CenterOfMass.z]

    if z[0] > 5:
        return None

    base = shape.Faces[z[0]]
    basepoint = base.Edges[0].Vertexes[0].Point
    plpoint = base.CenterOfMass
    # basenorm = base.normalAt(0.5, 0.5)

    # getting length and width
    vx = vec(base.Edges[0])
    vy = vec(base.Edges[1])
    if round(vx.Length) == round(vy.Length):
        vy = vec(base.Edges[2])

    # getting rotations
    rotZ = DraftVecUtils.angle(vx)
    rotY = DraftVecUtils.angle(vx, App.Vector(vx.x, vx.y, 0))
    rotX = DraftVecUtils.angle(vy, App.Vector(vy.x, vy.y, 0))

    # getting height
    vz = None
    rpi = round(math.pi / 2, precision())
    for i in range(1, 6):
        for e in shape.Faces[i].Edges:
            if basepoint in [e.Vertexes[0].Point, e.Vertexes[1].Point]:
                vtemp = vec(e)
                # print(vtemp)
                if round(vtemp.getAngle(vx), precision()) == rpi:
                    if round(vtemp.getAngle(vy), precision()) == rpi:
                        vz = vtemp

    if not vz:
        return None

    mat = App.Matrix()
    mat.move(plpoint)
    mat.rotateX(rotX)
    mat.rotateY(rotY)
    mat.rotateZ(rotZ)

    return [
        App.Placement(mat),
        round(vx.Length, precision()),
        round(vy.Length, precision()),
        round(vz.Length, precision()),
    ]


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Views command"""

import sys

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

UPDATEINTERVAL = 2000  # number of milliseconds between BIM Views window update
PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class BIM_Views:

    def GetResources(self):
        return {
            "Pixmap": "BIM_Views",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Views", "Views Manager"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_Views", "Shows or hides the views manager"),
            "Accel": "Ctrl+9",
        }

    def Activated(self):
        from PySide import QtCore, QtGui

        vm = findWidget()
        self.allItemsInTree = []
        self.oldData = [[], []]
        bimviewsbutton = None
        mw = FreeCADGui.getMainWindow()
        st = mw.statusBar()
        statuswidget = st.findChild(QtGui.QToolBar, "BIMStatusWidget")
        if statuswidget and hasattr(statuswidget, "bimviewsbutton"):
            bimviewsbutton = statuswidget.bimviewsbutton
        if vm:
            if vm.isVisible():
                vm.hide()
                if bimviewsbutton:
                    bimviewsbutton.setChecked(False)
                PARAMS.SetBool("RestoreBimViews", False)
            else:
                vm.show()
                if bimviewsbutton:
                    bimviewsbutton.setChecked(True)
                PARAMS.SetBool("RestoreBimViews", True)
                self.update()
        else:
            vm = QtGui.QDockWidget()

            # create the dialog
            self.dialog = FreeCADGui.PySideUic.loadUi(":/ui/dialogViews.ui")
            vm.setWidget(self.dialog)
            vm.tree = self.dialog.tree
            vm.viewtree = self.dialog.viewtree
            vm.closeEvent = self.onClose

            # set context menu
            self.dialog.tree.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)

            # set button
            self.dialog.menu = QtGui.QMenu()
            for button in [
                ("Active", translate("BIM", "Active")),
                ("AddLevel", translate("BIM", "New Level")),
                ("AddProxy", translate("BIM", "New Working Plane Proxy")),
                ("Delete", translate("BIM", "Delete")),
                ("Toggle", translate("BIM", "Toggle Visibility")),
                ("Isolate", translate("BIM", "Isolate")),
                ("SaveView", translate("BIM", "Save View Position")),
                ("Rename", translate("BIM", "Rename")),
            ]:
                action = QtGui.QAction(button[1])

                # Make the "Activate" button bold, as this is the default one
                if button[0] == "Active":
                    font = action.font()
                    font.setBold(True)
                    action.setFont(font)
                    action.setCheckable(True)

                self.dialog.menu.addAction(action)
                setattr(self.dialog, "button" + button[0], action)

            # # set button icons
            self.dialog.buttonAddLevel.setIcon(QtGui.QIcon(":/icons/Arch_Floor_Tree.svg"))
            self.dialog.buttonAddProxy.setIcon(QtGui.QIcon(":/icons/Draft_PlaneProxy.svg"))
            self.dialog.buttonDelete.setIcon(QtGui.QIcon(":/icons/delete.svg"))
            self.dialog.buttonToggle.setIcon(QtGui.QIcon(":/icons/dagViewVisible.svg"))
            self.dialog.buttonIsolate.setIcon(QtGui.QIcon(":/icons/Std_ShowSelection.svg"))
            self.dialog.buttonSaveView.setIcon(QtGui.QIcon(":/icons/Std_ViewScreenShot.svg"))
            self.dialog.buttonRename.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))

            # set tooltips
            self.dialog.buttonAddLevel.setToolTip(translate("BIM", "Creates a new level"))
            self.dialog.buttonAddProxy.setToolTip(
                translate("BIM", "Creates a new working plane proxy")
            )
            self.dialog.buttonDelete.setToolTip(translate("BIM", "Deletes the selected item"))
            self.dialog.buttonToggle.setToolTip(
                translate("BIM", "Toggles the visibility of selected items")
            )
            self.dialog.buttonIsolate.setToolTip(
                translate("BIM", "Turns all items off except the selected ones")
            )
            self.dialog.buttonSaveView.setToolTip(
                translate("BIM", "Saves the current camera position to the selected items")
            )
            self.dialog.buttonRename.setToolTip(translate("BIM", "Renames the selected item"))
            self.dialog.buttonActive.setToolTip(translate("BIM", "Activates the selected item"))

            # connect signals
            self.dialog.buttonAddLevel.triggered.connect(self.addLevel)
            self.dialog.buttonAddProxy.triggered.connect(self.addProxy)
            self.dialog.buttonDelete.triggered.connect(self.delete)
            self.dialog.buttonToggle.triggered.connect(self.toggle)
            self.dialog.buttonIsolate.triggered.connect(self.isolate)
            self.dialog.buttonSaveView.triggered.connect(self.saveView)
            self.dialog.buttonRename.triggered.connect(self.rename)
            self.dialog.buttonActive.triggered.connect(lambda: BIM_Views.activate(self.dialog))
            self.dialog.tree.itemClicked.connect(self.select)
            self.dialog.tree.itemDoubleClicked.connect(show)
            self.dialog.viewtree.itemDoubleClicked.connect(show)
            self.dialog.tree.itemChanged.connect(self.editObject)
            self.dialog.tree.customContextMenuRequested.connect(self.onContextMenu)
            # delay connecting after FreeCAD finishes setting up
            QtCore.QTimer.singleShot(UPDATEINTERVAL, self.connectDock)

            # set the dock widget
            area = PARAMS.GetInt("BimViewArea", 1)
            floating = PARAMS.GetBool("BimViewFloat", True)
            height = PARAMS.GetBool("BimViewWidth", 200)
            width = PARAMS.GetBool("BimViewHeight", 300)
            tabs = PARAMS.GetString("BimViewTabs", "")
            vm.setObjectName("BIM Views Manager")
            vm.setWindowTitle(translate("BIM", "BIM"))
            mw = FreeCADGui.getMainWindow()
            vm.setFloating(floating)
            vm.setGeometry(vm.x(), vm.y(), width, height)
            mw.addDockWidget(self.getDockArea(area), vm)
            if tabs:
                tabs = tabs.split("+")
                for tab in tabs:
                    dw = mw.findChild(QtGui.QDockWidget, tab)
                    if dw:
                        mw.tabifyDockWidget(dw, vm)
                        break

            # restore saved settings
            vm.tree.setColumnWidth(0, PARAMS.GetInt("ViewManagerColumnWidth", 100))
            vm.setFloating(PARAMS.GetBool("ViewManagerFloating", False))

            # check the status bar button
            if bimviewsbutton:
                bimviewsbutton.setChecked(True)
            PARAMS.SetBool("RestoreBimViews", True)

            self.update()

    def onClose(self, event):
        from PySide import QtGui

        st = FreeCADGui.getMainWindow().statusBar()
        statuswidget = st.findChild(QtGui.QToolBar, "BIMStatusWidget")
        if statuswidget and hasattr(statuswidget, "bimviewsbutton"):
            statuswidget.bimviewsbutton.setChecked(False)
        PARAMS.SetBool("RestoreBimViews", False)

    def connectDock(self):
        "watch for dock location"

        vm = findWidget()
        if vm:
            vm.dockLocationChanged.connect(self.onDockLocationChanged)

    def _treeToStringList(self, treeViewItems):
        "generates a (nested) string list representation of treeViewItems"

        def _toStringList(itm):
            children = []
            for i in range(itm.childCount()):
                children.append(_toStringList(itm.child(i)))
            return [itm.toolTip(0), itm.text(0), itm.text(1), children]

        return [_toStringList(itm) for itm in treeViewItems]

    def update(self, retrigger=True):
        "updates the view manager"

        from PySide import QtCore, QtGui
        import Draft

        vm = findWidget()
        if vm and vm.isVisible():
            if FreeCAD.isRestoring() or not FreeCAD.ActiveDocument:
                if vm.tree.state() != vm.tree.State.EditingState:
                    self.oldData[0] = []
                    vm.tree.clear()
                if vm.viewtree.state() != vm.viewtree.State.EditingState:
                    self.oldData[1] = []
                    vm.viewtree.clear()
            else:
                if vm.tree.state() != vm.tree.State.EditingState:
                    treeViewItems = []  # QTreeWidgetItem to Display in tree
                    lvHold = []
                    soloProxyHold = []
                    for obj in FreeCAD.ActiveDocument.Objects:
                        t = Draft.getType(obj)
                        if obj and (
                            t
                            in [
                                "Building",
                                "BuildingPart",
                                "IfcBuilding",
                                "IfcBuildingStorey",
                            ]
                        ):
                            if (
                                t in ["Building", "IfcBuilding"]
                                or getattr(obj, "IfcType", "") == "Building"
                            ):
                                building, _ = getTreeViewItem(obj)
                                subObjs = obj.Group
                                # find every levels belongs to the building
                                for subObj in subObjs:
                                    if Draft.getType(subObj) in [
                                        "BuildingPart",
                                        "Building Storey",
                                        "IfcBuildingStorey",
                                    ]:
                                        lv, lvH = getTreeViewItem(subObj)
                                        subSubObjs = subObj.Group
                                        # find every working plane proxy belongs to the level
                                        for subSubObj in subSubObjs:
                                            if Draft.getType(subSubObj) == "WorkingPlaneProxy":
                                                wp, _ = getTreeViewItem(subSubObj)
                                                lv.addChild(wp)
                                        lvHold.append((lv, lvH))
                                sortLvHold = sorted(lvHold, key=lambda x: x[1])
                                sortLvItems = [item[0] for item in sortLvHold]
                                for lvItem in sortLvItems:
                                    building.addChild(lvItem)
                                treeViewItems.append(building)
                                lvHold.clear()

                            if (
                                t in ["Building Storey", "IfcBuildingStorey"]
                                or getattr(obj, "IfcType", "") == "Building Storey"
                            ):
                                if (
                                    Draft.getType(getParent(obj)) in ["Building", "IfcBuilding"]
                                    or getattr(getParent(obj), "IfcType", "") == "Building"
                                ):
                                    continue
                                lv, lvH = getTreeViewItem(obj)
                                subObjs = obj.Group
                                # find every working plane proxy belongs to the level
                                for subObj in subObjs:
                                    if Draft.getType(subObj) == "WorkingPlaneProxy":
                                        wp, _ = getTreeViewItem(subObj)
                                        lv.addChild(wp)
                                lvHold.append((lv, lvH))
                        if obj and (t == "WorkingPlaneProxy"):
                            if obj.getParent() and obj.getParent().IfcType == "Building Storey":
                                continue
                            wp, _ = getTreeViewItem(obj)
                            soloProxyHold.append(wp)
                    sortLvHold = sorted(lvHold, key=lambda x: x[1])
                    sortLvItems = [item[0] for item in sortLvHold]
                    treeViewItems = treeViewItems + sortLvItems + soloProxyHold
                    new = self._treeToStringList(treeViewItems)
                    if new != self.oldData[0]:
                        self.oldData[0] = new
                        vm.tree.clear()
                        self.allItemsInTree.clear()
                        vm.tree.addTopLevelItems(treeViewItems)

                if vm.viewtree.state() != vm.viewtree.State.EditingState:
                    ficon = QtGui.QIcon.fromTheme("folder", QtGui.QIcon(":/icons/folder.svg"))
                    treeViewItems = []

                    views = self.getViews()
                    if views:
                        top = QtGui.QTreeWidgetItem([translate("BIM", "2D Views"), ""])
                        top.setIcon(0, ficon)
                        for v in views:
                            if hasattr(v, "Label"):
                                i = QtGui.QTreeWidgetItem([v.Label, ""])
                                if hasattr(v.ViewObject, "Icon"):
                                    i.setIcon(0, v.ViewObject.Icon)
                                i.setToolTip(0, v.Name)
                                top.addChild(i)
                        treeViewItems.append(top)

                    pages = self.getPages()
                    if pages:
                        top = QtGui.QTreeWidgetItem([translate("BIM", "Sheets"), ""])
                        top.setIcon(0, ficon)
                        for p in pages:
                            i = QtGui.QTreeWidgetItem([p.Label, ""])
                            if hasattr(p.ViewObject, "Icon"):
                                i.setIcon(0, p.ViewObject.Icon)
                            i.setToolTip(0, p.Name)
                            top.addChild(i)
                        treeViewItems.append(top)

                    new = self._treeToStringList(treeViewItems)
                    if new != self.oldData[1]:
                        self.oldData[1] = new
                        vm.viewtree.clear()
                        vm.viewtree.addTopLevelItems(treeViewItems)

            # We reuse the variable later on in "Isolate", to not traverse the tree once
            # again
            self.allItemsInTree = getAllItemsInTree(vm.tree)
            allItemsInTrees = self.allItemsInTree + getAllItemsInTree(vm.viewtree)

            if allItemsInTrees:
                # set TreeView Item selected if obj is selected
                objSelected = FreeCADGui.Selection.getSelection()
                objNameSelected = [obj.Name for obj in objSelected]
                objActive = FreeCADGui.ActiveDocument.ActiveView.getActiveObject("NativeIFC")
                if not objActive:
                    objActive = FreeCADGui.ActiveDocument.ActiveView.getActiveObject("Arch")

                default_background = allItemsInTrees[0].background(1)
                default_font = allItemsInTrees[0].font(1)
                for item in allItemsInTrees:
                    item.setSelected(item.toolTip(0) in objNameSelected)
                    if objActive and item.toolTip(0) == objActive.Name:
                        tparam = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/TreeView")
                        activeColor = tparam.GetUnsigned("TreeActiveColor", 0)
                        if activeColor:
                            r = ((activeColor >> 24) & 0xFF) / 255.0
                            g = ((activeColor >> 16) & 0xFF) / 255.0
                            b = ((activeColor >> 8) & 0xFF) / 255.0
                            activeColor = QtGui.QColor.fromRgbF(r, g, b)
                            item.setBackground(0, QtGui.QBrush(activeColor, QtCore.Qt.SolidPattern))
                            bold = QtGui.QFont()
                            bold.setBold(True)
                            item.setFont(0, bold)
                    else:
                        item.setBackground(0, default_background)
                        item.setFont(0, default_font)

        if retrigger:
            QtCore.QTimer.singleShot(UPDATEINTERVAL, self.update)

        # save state
        PARAMS.SetInt("ViewManagerColumnWidth", vm.tree.columnWidth(0))
        PARAMS.SetBool("ViewManagerFloating", vm.isFloating())

        # expand
        vm.tree.expandAll()
        vm.viewtree.expandAll()

    def select(self, item, column=None):
        "selects a doc object corresponding to an item"

        item.setSelected(True)
        name = item.toolTip(0)
        if name:
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                FreeCADGui.Selection.clearSelection()
                FreeCADGui.Selection.addSelection(obj)

    def addLevel(self):
        "adds a building part"

        import Arch

        FreeCAD.ActiveDocument.openTransaction("Create BuildingPart")
        obj = Arch.makeFloor()
        self.addToSelection(obj)
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()
        self.update(False)

    def addProxy(self):
        "adds a WP proxy"

        import Draft
        import WorkingPlane

        FreeCAD.ActiveDocument.openTransaction("Create WP Proxy")
        obj = Draft.makeWorkingPlaneProxy(WorkingPlane.get_working_plane().get_placement())
        self.addToSelection(obj)
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()
        self.update(False)

    def addToSelection(self, obj):
        "Adds the given object to the current selected item"

        import Draft
        from nativeifc import ifc_tools

        sel = FreeCADGui.Selection.getSelection()
        if len(sel) == 1:
            sel = sel[0]
            if hasattr(sel, "addObject"):
                sel.addObject(obj)
                return
            elif Draft.getType(sel).startswith("Ifc"):
                ifc_tools.aggregate(obj, sel)
            elif "Group" in sel.PropertiesList:
                g = sel.Group
                if obj not in g:
                    g.append(obj)
                sel.Group = g
                return

    def delete(self):
        "deletes the selected object"

        vm = findWidget()
        if vm:
            if vm.tree.selectedItems():
                FreeCAD.ActiveDocument.openTransaction("Delete")
                for item in vm.tree.selectedItems():
                    obj = FreeCAD.ActiveDocument.getObject(item.toolTip(0))
                    if obj:
                        FreeCAD.ActiveDocument.removeObject(obj.Name)
                FreeCAD.ActiveDocument.commitTransaction()
                FreeCAD.ActiveDocument.recompute()
                self.update(False)

    def rename(self):
        "renames the selected object"

        vm = findWidget()
        if vm:
            if vm.tree.selectedItems():
                if vm.tree.selectedItems():
                    item = vm.tree.selectedItems()[-1]
                    vm.tree.editItem(item, 0)

    @staticmethod
    def activate(dialog=None):
        from draftutils.gui_utils import toggle_working_plane

        vm = findWidget()
        if vm:
            if vm.tree.selectedItems():
                item = vm.tree.selectedItems()[-1]
                obj = FreeCAD.ActiveDocument.getObject(item.toolTip(0))
                if obj:
                    toggle_working_plane(obj, None, restore=True, dialog=dialog)
                    FreeCADGui.Selection.clearSelection()

    def editObject(self, item, column):
        "renames or edit height of the actual object"

        obj = FreeCAD.ActiveDocument.getObject(item.toolTip(0))
        if obj:
            if column == 0:
                obj.Label = item.text(column)
            if column == 1:
                obj.Placement.Base.z = FreeCAD.Units.parseQuantity(item.text(column))

    def toggle(self):
        "toggle selected item on/off"

        vm = findWidget()
        if vm:
            for item in vm.tree.selectedItems():
                obj = FreeCAD.ActiveDocument.getObject(item.toolTip(0))
                if obj:
                    obj.ViewObject.Visibility = not (obj.ViewObject.Visibility)
            FreeCAD.ActiveDocument.recompute()

    def _isAncestor(self, ancestor_item, child_item):
        current = child_item.parent()
        while current is not None:
            if current == ancestor_item:
                return True
            current = current.parent()
        return False

    def isolate(self):
        import Draft

        """
        Isolate the currently selected items in the tree view.

        This function first makes all items in the tree visible to ensure a clean slate.
        Then, it hides all items that are not currently selected by the user in the GUI tree view.
        As a result, only the selected items remain visible in the 3D view, effectively isolating them.

        Assumes that `self.allItemsInTree` is a list of all QTreeWidgetItems in the tree.
        """

        # Iterate through all of the items and show them beforehand if they were hidden
        # so we can "reset" the tree state before the real processing
        for item in self.allItemsInTree:
            toolTip = item.toolTip(0)
            obj = FreeCAD.ActiveDocument.getObject(toolTip)
            if obj:
                # We switch visibility to be sure we will show childs of other childs
                # beforehand, as the Visibility may not be propagated.
                obj.ViewObject.Visibility = False
                obj.ViewObject.Visibility = True

        vm = findWidget()
        if vm:
            selectedItems = vm.tree.selectedItems()
            checkAncestors = False
            # We can get a scenario where user has just selected only Building
            # so we don't want to hide any of it's children, so just check if that's
            # the case so we will know whether we should process items further or not
            if len(selectedItems) == 1:
                toolTip = selectedItems[0].toolTip(0)
                obj = FreeCAD.ActiveDocument.getObject(toolTip)
                t = Draft.getType(obj)
                if obj and getattr(obj, "IfcType", "") == "Building":
                    checkAncestors = True

            for item in self.allItemsInTree:
                toolTip = item.toolTip(0)
                obj = FreeCAD.ActiveDocument.getObject(toolTip)
                if obj:
                    if item not in selectedItems and not (
                        checkAncestors and self._isAncestor(selectedItems[0], item)
                    ):
                        obj.ViewObject.Visibility = False
                    else:
                        obj.ViewObject.Visibility = True

    def saveView(self):
        "save the current camera angle to the selected item"

        vm = findWidget()
        if vm:
            for item in vm.tree.selectedItems():
                obj = FreeCAD.ActiveDocument.getObject(item.toolTip(0))
                if obj:
                    if hasattr(obj.ViewObject.Proxy, "writeCamera"):
                        obj.ViewObject.Proxy.writeCamera()
        FreeCAD.ActiveDocument.recompute()

    def onDockLocationChanged(self, area):
        """Saves dock widget size and location"""
        if hasattr(area, "value"):  # To support Qt5.15
            PARAMS.SetInt("BimViewArea", area.value)
        else:
            PARAMS.SetInt("BimViewArea", int(area))
        mw = FreeCADGui.getMainWindow()
        vm = findWidget()
        if vm:
            PARAMS.SetBool("BimViewFloat", vm.isFloating())
            PARAMS.SetInt("BimViewWidth", vm.width())
            PARAMS.SetInt("BimViewHeight", vm.height())
            tabs = "+".join([o.objectName() for o in mw.tabifiedDockWidgets(vm)])
            PARAMS.SetString("BimViewTabs", tabs)

    def getDockArea(self, area):
        """Turns an int into a qt dock area"""

        from PySide import QtCore

        if area == 1:
            return QtCore.Qt.LeftDockWidgetArea
        elif area == 4:
            return QtCore.Qt.TopDockWidgetArea
        elif area == 8:
            return QtCore.Qt.BottomDockWidgetArea
        else:
            return QtCore.Qt.RightDockWidgetArea

    def onContextMenu(self, pos):
        """Fires the context menu"""
        import Draft

        self.dialog.buttonAddProxy.setEnabled(True)
        selobj = self.dialog.tree.currentItem()
        if selobj:
            selobj = FreeCAD.ActiveDocument.getObject(selobj.toolTip(0))
            if selobj:
                if Draft.getType(selobj).startswith("Ifc"):
                    self.dialog.buttonAddProxy.setEnabled(False)
                if FreeCADGui.ActiveDocument.ActiveView.getActiveObject("Arch") == selobj:
                    self.dialog.buttonActive.setChecked(True)
                else:
                    self.dialog.buttonActive.setChecked(False)
        self.dialog.menu.exec_(self.dialog.tree.mapToGlobal(pos))

    def getViews(self):
        """Returns a list of 2D views"""
        import Draft

        views = []
        for p in self.getPages():
            for v in p.Views:
                if getattr(v, "Source", None):
                    views.append(v.Source)
        bps = [o for o in FreeCAD.ActiveDocument.Objects if Draft.getType(o) == "BuildingPart"]
        for v in [o for o in bps if isView(o)]:
            if v not in views:
                views.append(v)
        return views

    def getPages(self):
        """Returns a list of TD pages"""
        return [o for o in FreeCAD.ActiveDocument.Objects if o.isDerivedFrom("TechDraw::DrawPage")]


# These functions need to be localized outside the command class, as they are used outside this module


def findWidget():
    "finds the manager widget, if present"

    from PySide import QtGui

    mw = FreeCADGui.getMainWindow()
    vm = mw.findChild(QtGui.QDockWidget, "BIM Views Manager")
    if vm:
        return vm
    return None


def show(item, column=None):
    "item has been double-clicked"
    import Draft

    obj = None
    vm = findWidget()
    if isinstance(item, str) or ((sys.version_info.major < 3) and isinstance(item, unicode)):
        # called from Python code
        obj = FreeCAD.ActiveDocument.getObject(item)
    else:
        # called from GUI
        if column == 1:
            # user clicked the level field
            if vm:
                vm.tree.editItem(item, column)
                return
        else:
            # TODO find a way to not edit the object name
            obj = FreeCAD.ActiveDocument.getObject(item.toolTip(0))
    if obj:
        FreeCADGui.Selection.clearSelection()
        FreeCADGui.Selection.addSelection(obj)
        vparam = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View")
        if obj.isDerivedFrom("TechDraw::DrawPage"):

            # case 1: the object is a TD page. We switch to it simply
            obj.ViewObject.Visibility = True
        elif isView(obj):

            # case 2: the object is a 2D view
            ssel = [obj] + obj.OutListRecursive
            FreeCADGui.Selection.clearSelection()
            for o in ssel:
                o.ViewObject.Visibility = True
                FreeCADGui.Selection.addSelection(o)
            if not hasattr(FreeCADGui.ActiveDocument.ActiveView, "getSceneGraph"):
                # Find first 3d view and switch to it
                for w in FreeCADGui.getMainWindow().getWindows():
                    if hasattr(w, "getSceneGraph"):
                        FreeCADGui.getMainWindow().setActiveWindow(w)
                        break
            FreeCADGui.runCommand("Std_OrthographicCamera")
            FreeCADGui.ActiveDocument.ActiveView.viewTop()
            FreeCADGui.SendMsgToActiveView("ViewSelection")
            FreeCADGui.ActiveDocument.ActiveView.viewTop()
            FreeCADGui.Selection.clearSelection()
            FreeCADGui.Selection.addSelection(obj)
            if PARAMS.GetBool("BimViewsSwitchBackground", False):
                vparam.SetBool("Simple", True)
                vparam.SetBool("Gradient", False)
                vparam.SetBool("RadialGradient", False)
        else:
            # case 3: This is maybe a BuildingPart. Place the WP on it")
            type = Draft.getType(obj)
            if type == "BuildingPart" or type == "IfcBuildingStorey":
                BIM_Views.activate()

    if vm:
        # store the last double-clicked item for the BIM WPView command
        if isinstance(item, str) or ((sys.version_info.major < 3) and isinstance(item, unicode)):
            vm.lastSelected = item
        else:
            vm.lastSelected = item.toolTip(0)


def isView(obj):
    """Returns true if this object is used as Source of a view on a TD page"""

    for p in obj.InList:
        if p.isDerivedFrom("TechDraw::DrawView"):
            if hasattr(p, "Source"):
                if p.Source == obj:
                    return True
    if getattr(obj, "DrawingView", False):
        return True
    if getattr(obj, "IfcType", None) == "Annotation":
        if getattr(obj, "ObjectType", "").upper() == "DRAWING":
            return True
    if getattr(obj, "Class", None) == "IfcAnnotation":
        if getattr(obj, "ObjectType", "").upper() == "DRAWING":
            return True
    return False


def getTreeViewItem(obj):
    """
    from FreeCAD object make the TreeWidgetItem including icon Label and LevelHeight
    and also make a level height in number to sort the order after
    """
    from PySide import QtCore, QtGui

    z = obj.Placement.Base.z
    lvHStr = FreeCAD.Units.Quantity(z, FreeCAD.Units.Length).UserString
    if z == 0:
        # override with Elevation property if available
        if hasattr(obj, "Elevation"):
            z = obj.Elevation.Value
            lvHStr = obj.Elevation.UserString
    it = QtGui.QTreeWidgetItem([obj.Label, lvHStr])
    it.setFlags(it.flags() | QtCore.Qt.ItemIsEditable)
    it.setToolTip(0, obj.Name)
    if obj.ViewObject:
        if hasattr(obj.ViewObject, "Proxy") and hasattr(obj.ViewObject.Proxy, "getIcon"):
            it.setIcon(0, QtGui.QIcon(obj.ViewObject.Proxy.getIcon()))
    return (it, z)


def getAllItemsInTree(tree_widget):
    "return list of all items in QtreeWidget"

    def get_child_items(parent_item):
        child_items = []
        # get how many sub items
        child_count = parent_item.childCount()
        for j in range(child_count):
            child_item = parent_item.child(j)
            child_items.append(child_item)
            child_items.extend(get_child_items(child_item))

        return child_items

    all_items = []
    # get top level items
    top_level_item_count = tree_widget.topLevelItemCount()
    for i in range(top_level_item_count):
        top_level_item = tree_widget.topLevelItem(i)
        all_items.append(top_level_item)
        # iterate sub-items
        all_items.extend(get_child_items(top_level_item))

    return all_items


def getParent(obj):
    "return the first parent of this object"

    if obj.getParent():
        return obj.getParent()
    else:
        for parent in obj.InList:
            if hasattr(parent, "Group") and obj in parent.Group:
                return parent


FreeCADGui.addCommand("BIM_Views", BIM_Views())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Yorik van Havre <yorik@uncreated.net>              *
# *   Reimplemented as dressup in 2025  phaseloop <phaseloop@protonmail.com *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


import FreeCAD
import Path
import PathScripts.PathUtils as PathUtils
import Path.Dressup.Utils as PathDressup
import random
from PySide.QtCore import QT_TRANSLATE_NOOP

__doc__ = """CAM Array dressup"""

translate = FreeCAD.Qt.translate


class DressupArray:
    def __init__(self, obj, base, job):
        obj.addProperty(
            "App::PropertyLink",
            "Base",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The base toolpath to modify"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "Type",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Pattern method"),
        )
        obj.addProperty(
            "App::PropertyVectorDistance",
            "Offset",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The spacing between the array copies in linear pattern",
            ),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "CopiesX",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property", "The number of copies in X-direction in linear pattern"
            ),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "CopiesY",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property", "The number of copies in Y-direction in linear pattern"
            ),
        )
        obj.addProperty(
            "App::PropertyAngle",
            "Angle",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Total angle in polar pattern"),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "Copies",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property", "The number of copies in linear 1D and polar pattern"
            ),
        )
        obj.addProperty(
            "App::PropertyVector",
            "Centre",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The centre of rotation in polar pattern"),
        )
        obj.addProperty(
            "App::PropertyBool",
            "SwapDirection",
            "Path",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Make copies in X-direction before Y in linear 2D pattern",
            ),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "JitterPercent",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Percent of copies to randomly offset"),
        )
        obj.addProperty(
            "App::PropertyVectorDistance",
            "JitterMagnitude",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Maximum random offset of copies"),
        )
        obj.addProperty(
            "App::PropertyInteger",
            "JitterSeed",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "Seed value for jitter randomness"),
        )

        self.obj = obj
        obj.Base = base

        # assigning array tells the type of possible enum choices
        obj.Type = ["Linear1D", "Linear2D", "Polar"]
        # assign value
        obj.Type = "Linear1D"

        obj.Copies = 0
        obj.JitterPercent = 0

        self.setEditorModes(obj)
        obj.Proxy = self

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def setEditorModes(self, obj):
        if obj.Type == "Linear1D":
            angleMode = centreMode = copiesXMode = copiesYMode = swapDirectionMode = 2
            copiesMode = offsetMode = 0
        elif obj.Type == "Linear2D":
            angleMode = copiesMode = centreMode = 2
            copiesXMode = copiesYMode = offsetMode = swapDirectionMode = 0
        elif obj.Type == "Polar":
            angleMode = copiesMode = centreMode = 0
            copiesXMode = copiesYMode = offsetMode = swapDirectionMode = 2

        obj.setEditorMode("Angle", angleMode)
        obj.setEditorMode("Copies", copiesMode)
        obj.setEditorMode("Centre", centreMode)
        obj.setEditorMode("CopiesX", copiesXMode)
        obj.setEditorMode("CopiesY", copiesYMode)
        obj.setEditorMode("Offset", offsetMode)
        obj.setEditorMode("SwapDirection", swapDirectionMode)
        obj.setEditorMode("JitterPercent", 0)
        obj.setEditorMode("JitterMagnitude", 0)
        obj.setEditorMode("JitterSeed", 0)

    def onChanged(self, obj, prop):
        if prop == "Type":
            self.setEditorModes(obj)
        if prop == "Path" and obj.ViewObject:
            obj.ViewObject.signalChangeIcon()

    def onDocumentRestored(self, obj):
        """onDocumentRestored(obj) ... Called automatically when document is restored."""

        # Before commit 24c32c9 array dressups did have their own CoolantMode
        # and ToolController which they copied from the base operation. We need
        # to remove those properties when loading a file that was created before
        # that commit. Otherwise the wrong or no CoolantMode/ToolController will
        # be used.

        if hasattr(obj, "CoolantMode"):
            obj.removeProperty("CoolantMode")
            FreeCAD.Console.PrintWarning(
                translate(
                    "DressupArray",
                    "Removing CoolantMode property from {} as base operation's CoolantMode is now used.",
                ).format(obj.Name)
                + "\n"
            )

        if hasattr(obj, "ToolController"):
            obj.removeProperty("ToolController")
            FreeCAD.Console.PrintWarning(
                translate(
                    "DressupArray",
                    "Removing ToolController property from {} as base operation's ToolController is now used.",
                ).format(obj.Name)
                + "\n"
            )

        self.obj = obj
        self.setEditorModes(obj)

    def onDelete(self, obj, args):
        if obj.Base:
            job = PathUtils.findParentJob(obj)
            if job:
                job.Proxy.addOperation(obj.Base, obj)
            if obj.Base.ViewObject:
                obj.Base.ViewObject.Visibility = True
            obj.Base = None
        return True

    def execute(self, obj):

        if not obj.Base or not obj.Base.isDerivedFrom("Path::Feature") or not obj.Base.Path:
            Path.Log.error(translate("PathArray", "Base is empty or an invalid object."))
            return None

        # Do not generate paths and clear current Path data if operation not active
        if not PathDressup.baseOp(obj.Base).Active:
            if obj.Path:
                obj.Path = Path.Path()
            return

        # use seed if specified, otherwise default to object name for consistency during recomputes
        seed = obj.JitterSeed or obj.Name

        pa = PathArray(
            obj.Base,
            obj.Type,
            obj.Copies,
            obj.Offset,
            obj.CopiesX,
            obj.CopiesY,
            obj.Angle,
            obj.Centre,
            obj.SwapDirection,
            obj.JitterMagnitude,
            obj.JitterPercent,
            seed,
        )

        obj.Path = pa.getPath()


class PathArray:
    """class PathArray ...
    This class receives one or more base operations and repeats those operations
    at set intervals based upon array type requested and the related settings for that type."""

    def __init__(
        self,
        base,
        arrayType,
        copies,
        offsetVector,
        copiesX,
        copiesY,
        angle,
        centre,
        swapDirection,
        jitterMagnitude=FreeCAD.Vector(0, 0, 0),
        jitterPercent=0,
        seed="FreeCAD",
    ):
        self.base = base
        self.arrayType = arrayType  # ['Linear1D', 'Linear2D', 'Polar']
        self.copies = copies
        self.offsetVector = offsetVector
        self.copiesX = copiesX
        self.copiesY = copiesY
        self.angle = angle
        self.centre = centre
        self.swapDirection = swapDirection
        self.jitterMagnitude = jitterMagnitude
        self.jitterPercent = jitterPercent
        self.seed = seed

    # Private method
    def _calculateJitter(self, pos):
        """_calculateJitter(pos) ...
        Returns the position argument with a random vector shift applied."""
        if self.jitterPercent == 0:
            pass
        elif random.randint(0, 100) < self.jitterPercent:
            pos.x = pos.x + random.uniform(-self.jitterMagnitude.x, self.jitterMagnitude.x)
            pos.y = pos.y + random.uniform(-self.jitterMagnitude.y, self.jitterMagnitude.y)
            pos.z = pos.z + random.uniform(-self.jitterMagnitude.z, self.jitterMagnitude.z)
        return pos

    # Public method
    def getPath(self):
        """getPath() ... Call this method on an instance of the class to generate and return
        path data for the requested path array."""

        if self.base is None:
            Path.Log.error(translate("PathArray", "No base objects for PathArray."))
            return None

        base = self.base

        # build copies
        # initially output contains original base path, copies are added on top of that
        output = PathUtils.getPathWithPlacement(base).toGCode()

        random.seed(self.seed)

        if self.arrayType == "Linear1D":
            for i in range(self.copies):
                pos = FreeCAD.Vector(
                    self.offsetVector.x * (i + 1),
                    self.offsetVector.y * (i + 1),
                    self.offsetVector.z * (i + 1),
                )
                pos = self._calculateJitter(pos)

                pl = FreeCAD.Placement()
                pl.move(pos)
                np = Path.Path(
                    [cm.transform(pl) for cm in PathUtils.getPathWithPlacement(base).Commands]
                )

                output += np.toGCode()

        elif self.arrayType == "Linear2D":
            if self.swapDirection:
                for i in range(self.copiesY + 1):
                    for j in range(self.copiesX + 1):
                        if (i % 2) == 0:
                            pos = FreeCAD.Vector(
                                self.offsetVector.x * j,
                                self.offsetVector.y * i,
                                self.offsetVector.z * i,
                            )
                        else:
                            pos = FreeCAD.Vector(
                                self.offsetVector.x * (self.copiesX - j),
                                self.offsetVector.y * i,
                                self.offsetVector.z * i,
                            )
                        pos = self._calculateJitter(pos)

                        pl = FreeCAD.Placement()
                        # do not process the index 0,0. It will be processed by the base Paths themselves
                        if not (i == 0 and j == 0):
                            pl.move(pos)
                            np = Path.Path(
                                [
                                    cm.transform(pl)
                                    for cm in PathUtils.getPathWithPlacement(base).Commands
                                ]
                            )
                            output += np.toGCode()
            else:
                for i in range(self.copiesX + 1):
                    for j in range(self.copiesY + 1):
                        if (i % 2) == 0:
                            pos = FreeCAD.Vector(
                                self.offsetVector.x * i,
                                self.offsetVector.y * j,
                                self.offsetVector.z * i,
                            )
                        else:
                            pos = FreeCAD.Vector(
                                self.offsetVector.x * i,
                                self.offsetVector.y * (self.copiesY - j),
                                self.offsetVector.z * i,
                            )
                        pos = self._calculateJitter(pos)

                        pl = FreeCAD.Placement()
                        # do not process the index 0,0. It will be processed by the base Paths themselves
                        if not (i == 0 and j == 0):
                            pl.move(pos)
                            np = Path.Path(
                                [
                                    cm.transform(pl)
                                    for cm in PathUtils.getPathWithPlacement(base).Commands
                                ]
                            )
                            output += np.toGCode()
            # Eif
        else:
            for i in range(self.copies):
                ang = 360
                if self.copies > 0:
                    ang = self.angle / self.copies * (1 + i)

                pl = FreeCAD.Placement()
                pl.rotate(self.centre, FreeCAD.Vector(0, 0, 1), ang)
                np = PathUtils.applyPlacementToPath(pl, PathUtils.getPathWithPlacement(base))
                output += np.toGCode()

        # return output
        return Path.Path(output)


def Create(base, name="DressupArray"):
    """Create(base, name='DressupPathBoundary') ... creates a dressup array."""

    if not base.isDerivedFrom("Path::Feature"):
        Path.Log.error(translate("CAM_DressupArray", "The selected object is not a path") + "\n")
        return None

    obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    job = PathUtils.findParentJob(base)
    obj.Proxy = DressupArray(obj, base, job)
    job.Proxy.addOperation(obj, base, True)
    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import json
from typing import Optional, Mapping
import FreeCAD
import Path
from ...camassets import cam_assets
from ..mixins import RotaryToolBitMixin
from ..models.base import ToolBit
from ...assets.serializer import AssetSerializer
from ...assets.uri import AssetUri
from ...assets.asset import Asset

SHAPEMAP = {
    "ballend": "Ballnose",
    "endmill": "Cylindrical",
    "taperedballnose": "Ballnose",
    "v-bit": "Conical",
    "vbit": "Conical",
    "chamfer": "Snubnose",
}
SHAPEMAP_REVERSE = {
    "Ballnose": "ballend",  # Default to ballend when deserializing Ballnose
    "Cylindrical": "endmill",
    "Conical": "v-bit",
    "Snubnose": "chamfer",
}

tooltemplate = {
    "units": "metric",
    "shape": "Cylindrical",
    "length": 10,
    "diameter": 3,
    "description": "",
}


class CamoticsToolBitSerializer(AssetSerializer):
    for_class = ToolBit
    extensions = tuple()  # Camotics does not have tool files; tools are rows in tool tables
    mime_type = "application/json"
    can_import = False
    can_export = False

    @classmethod
    def get_label(cls) -> str:
        return FreeCAD.Qt.translate("CAM", "Camotics Tool")

    @classmethod
    def extract_dependencies(cls, data: bytes) -> list[AssetUri]:
        return []

    @classmethod
    def serialize(cls, asset: Asset) -> bytes:
        assert isinstance(asset, ToolBit)
        if not isinstance(asset, RotaryToolBitMixin):
            lbl = asset.label
            name = asset.get_shape_name()
            Path.Log.info(
                f"Skipping export of toolbit {lbl} ({name}) because it is not a rotary tool."
            )
            return b"{}"
        toolitem = tooltemplate.copy()
        toolitem["diameter"] = asset.get_diameter().Value or 2
        toolitem["description"] = asset.label
        toolitem["length"] = asset.get_length().Value or 10
        toolitem["shape"] = SHAPEMAP.get(asset.get_shape_name(), "Cylindrical")
        return json.dumps(toolitem).encode("ascii", "ignore")

    @classmethod
    def deserialize(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> ToolBit:
        # Create an instance of the ToolBitShape class
        attrs: dict = json.loads(data.decode("ascii", "ignore"))
        shape = cam_assets.get("toolbitshape://endmill")

        # Create an instance of the ToolBit class
        bit = ToolBit.from_shape_id(shape.get_id())
        bit.label = attrs["description"]

        if not isinstance(bit, RotaryToolBitMixin):
            raise NotImplementedError(
                f"Only export of rotary tools is supported ({bit.label} ({bit.id})"
            )

        bit.set_diameter(FreeCAD.Units.Quantity(float(attrs["diameter"]), "mm"))
        bit.set_length(FreeCAD.Units.Quantity(float(attrs["length"]), "mm"))
        return bit



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2021 sliptonic shopinthewoods@gmail.com                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "CAM Machine State"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Dataclass to implement a machinestate tracker"
__contributors__ = ""

import Path
import FreeCAD
import Path

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class MachineState:
    def __init__(self):
        self.WCSLIST = [
            "G53",
            "G54",
            "G55",
            "G56",
            "G57",
            "G58",
            "G59",
            "G59.1",
            "G59.2",
            "G59.3",
            "G59.4",
            "G59.5",
            "G59.6",
            "G59.7",
            "G59.8",
            "G59.9",
        ]

        self.X = 0.0  #: float = field(default=0)
        self.Y = 0.0  #: float = field(default=0)
        self.Z = 0.0  #: float = field(default=0)
        self.A = 0.0  #: float = field(default=0)
        self.B = 0.0  #: float = field(default=0)
        self.C = 0.0  #: float = field(default=0)
        self.F = 0.0  #: float = field(default=None)
        self.Coolant = False  #: bool = field(default=False)
        self.WCS = "G54"  #: str = field(default="G54")
        self.Spindle = "off"  #: str = field(default="off")
        self.S = 0  #: int = field(default=0)
        self.T = None  #: int = field(default=None)

    def addCommand(self, command):
        """Processes a command and updates the internal state of the machine. Returns true if the command has alterned the machine state"""
        oldstate = self.getState()
        if command.Name == "M6":
            self.T = int(command.Parameters["T"])
            return not oldstate == self.getState()

        if command.Name in ["M3", "M4"]:
            self.S = command.Parameters["S"]
            self.Spindle = "CW" if command.Name == "M3" else "CCW"
            return not oldstate == self.getState()

        if command.Name in ["M2", "M5"]:
            self.S = 0
            self.Spindle = "off"
            return not oldstate == self.getState()

        if command.Name in self.WCSLIST:
            self.WCS = command.Name
            return not oldstate == self.getState()

        if command.Name in Path.Geom.CmdMoveDrill:
            oldZ = self.Z
            for p in command.Parameters:
                self.__setattr__(p, command.Parameters[p])
            self.__setattr__("Z", oldZ)
            return not oldstate == self.getState()

        for p in command.Parameters:
            self.__setattr__(p, command.Parameters[p])

        return not oldstate == self.getState()

    def getState(self):
        """
        Returns a dictionary of the current machine state
        """
        state = {}
        state["X"] = self.X
        state["Y"] = self.Y
        state["Z"] = self.Z
        state["A"] = self.A
        state["B"] = self.B
        state["C"] = self.C
        state["F"] = self.F
        state["Coolant"] = self.Coolant
        state["WCS"] = self.WCS
        state["Spindle"] = self.Spindle
        state["S"] = self.S
        state["T"] = self.T

        return state

    def getPosition(self):
        """
        Returns a vector of the current machine position
        """

        # This is technical debt.  The actual position may include a rotation
        # component as well.  We should probably be returning a placement
        return FreeCAD.Vector(self.X, self.Y, self.Z)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


import Path
import numpy

__title__ = "Drilling Toolpath Generator"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Generates the drilling toolpath for a single spotshape"


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


def generate(
    edge,
    dwelltime=0.0,
    peckdepth=0.0,
    repeat=1,
    retractheight=None,
    chipBreak=False,
    feedRetract=False,
):
    """
    Generates Gcode for drilling a single hole.

    Takes as input an edge. It assumes the edge is trivial with just two vectors.
    The edge must be aligned with the Z axes (Vector(0,0,1)) or it is an error.

    The first vertex of the edge will be the startpoint
    The second vertex of the edge will be the endpoint.
    All other vertices are ignored.

    additionally, you can pass in a dwelltime, peckdepth, and repeat value.

    These will result in appropriate G81,G82, and G83 codes.

    If chipBreak is True, the generator will produce G73 cycles instead of G83.
    Chipbreaking cycles produce very small retracts to break the chip rather than
    full retracts to clear chips from the hole.
    http://linuxcnc.org/docs/html/gcode/g-code.html#gcode:g73

    If feedRetract is True, the generator will produce G85 cycles which retract
    the tool at the specified feed rate instead of performing a rapid move.
    This is useful for boring or reaming operations. Peck or dwell cannot be used with feed retract.
    http://linuxcnc.org/docs/html/gcode/g-code.html#gcode:g85

    """
    startPoint = edge.Vertexes[0].Point
    endPoint = edge.Vertexes[1].Point

    Path.Log.debug(startPoint)
    Path.Log.debug(endPoint)

    Path.Log.debug(numpy.isclose(startPoint.sub(endPoint).x, 0, rtol=1e-05, atol=1e-06))
    Path.Log.debug(numpy.isclose(startPoint.sub(endPoint).y, 0, rtol=1e-05, atol=1e-06))
    Path.Log.debug(endPoint)

    if dwelltime > 0.0 and peckdepth > 0.0:
        raise ValueError("Peck and Dwell cannot be used together")

    if dwelltime > 0.0 and feedRetract:
        raise ValueError("Dwell and feed retract cannot be used together")

    if peckdepth > 0.0 and feedRetract:
        raise ValueError("Peck and feed retract cannot be used together")

    if repeat < 1:
        raise ValueError("repeat must be 1 or greater")

    if not type(repeat) is int:
        raise ValueError("repeat value must be an integer")

    if not type(peckdepth) is float:
        raise ValueError("peckdepth must be a float")

    if not type(dwelltime) is float:
        raise ValueError("dwelltime must be a float")

    if retractheight is not None and not type(retractheight) is float:
        raise ValueError("retractheight must be a float")

    if not (
        numpy.isclose(startPoint.sub(endPoint).x, 0, rtol=1e-05, atol=1e-06)
        and (numpy.isclose(startPoint.sub(endPoint).y, 0, rtol=1e-05, atol=1e-06))
    ):
        raise ValueError("edge is not aligned with Z axis")

    if startPoint.z < endPoint.z:
        raise ValueError("start point is below end point")

    cmdParams = {}
    cmdParams["X"] = startPoint.x
    cmdParams["Y"] = startPoint.y
    cmdParams["Z"] = endPoint.z
    cmdParams["R"] = retractheight if retractheight is not None else startPoint.z

    if repeat < 1:
        raise ValueError("repeat must be 1 or greater")

    if not type(repeat) is int:
        raise ValueError("repeat value must be an integer")

    if repeat > 1:
        cmdParams["L"] = repeat

    if feedRetract:
        cmd = "G85"
    elif peckdepth == 0.0:
        if dwelltime > 0.0:
            cmd = "G82"
            cmdParams["P"] = dwelltime
        else:
            cmd = "G81"
    else:
        cmd = "G73" if chipBreak else "G83"
        cmdParams["Q"] = peckdepth

    return [Path.Command(cmd, cmdParams)]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Op.Base as PathOp
import Path.Op.EngraveBase as PathEngraveBase
import PathScripts.PathUtils as PathUtils

from PySide.QtCore import QT_TRANSLATE_NOOP

__doc__ = "Class and implementation of CAM Engrave operation"

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")


class ObjectEngrave(PathEngraveBase.ObjectOp):
    """Proxy class for Engrave operation."""

    def __init__(self, obj, name, parentJob):
        super(ObjectEngrave, self).__init__(obj, name, parentJob)
        self.wires = []

    def opFeatures(self, obj):
        """opFeatures(obj) ... return all standard features and edges based geometries"""
        return (
            PathOp.FeatureTool
            | PathOp.FeatureDepths
            | PathOp.FeatureHeights
            | PathOp.FeatureStepDown
            | PathOp.FeatureBaseEdges
            | PathOp.FeatureCoolant
        )

    def setupAdditionalProperties(self, obj):
        if not hasattr(obj, "BaseShapes"):
            obj.addProperty(
                "App::PropertyLinkList",
                "BaseShapes",
                "Path",
                QT_TRANSLATE_NOOP("App::Property", "Additional base objects to be engraved"),
            )
        obj.setEditorMode("BaseShapes", 2)  # hide
        if not hasattr(obj, "BaseObject"):
            obj.addProperty(
                "App::PropertyLink",
                "BaseObject",
                "Path",
                QT_TRANSLATE_NOOP("App::Property", "Additional base objects to be engraved"),
            )
        obj.setEditorMode("BaseObject", 2)  # hide

    def initOperation(self, obj):
        """initOperation(obj) ... create engraving specific properties."""
        obj.addProperty(
            "App::PropertyInteger",
            "StartVertex",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The vertex index to start the toolpath from"),
        )
        self.setupAdditionalProperties(obj)

    def opOnDocumentRestored(self, obj):
        # upgrade ...
        self.setupAdditionalProperties(obj)

    def opExecute(self, obj):
        """opExecute(obj) ... process engraving operation"""
        Path.Log.track()

        jobshapes = []

        if obj.Base:
            # user has selected specific subelements
            Path.Log.track(len(obj.Base))
            for base, subs in obj.Base:
                edges = []
                wires = []
                for feature in subs:
                    sub = base.Shape.getElement(feature)
                    if isinstance(sub, Part.Edge):
                        edges.append(sub)
                    elif sub.Wires:
                        wires.extend(sub.Wires)
                    else:
                        wires.append(Part.Wire(sub.Edges))

                for sortedEdges in Part.sortEdges(edges):
                    wires.append(Part.Wire(sortedEdges))

                jobshapes.append(Part.makeCompound(wires))

        elif obj.BaseShapes:
            # user added specific shapes
            jobshapes.extend([base.Shape for base in obj.BaseShapes])
        else:
            # process all objects in Job.Model.Group
            Path.Log.track(self.model)
            for base in self.model:
                Path.Log.track(base.Label)
                if base.isDerivedFrom("Part::Part2DObject"):
                    jobshapes.append(base.Shape)
                elif base.isDerivedFrom("Sketcher::SketchObject"):
                    jobshapes.append(base.Shape)
                elif hasattr(base, "ArrayType"):
                    jobshapes.append(base.Shape)

        if jobshapes:
            Path.Log.debug("processing {} jobshapes".format(len(jobshapes)))
            wires = []
            for shape in jobshapes:
                if isinstance(shape, Part.Edge):
                    shapeWires = [Part.Wire(shape)]
                else:
                    shapeWires = shape.Wires
                Path.Log.debug("jobshape has {} edges".format(len(shape.Edges)))
                self.commandlist.append(
                    Path.Command("G0", {"Z": obj.ClearanceHeight.Value, "F": self.vertRapid})
                )
                self.buildpathocc(obj, shapeWires, self.getZValues(obj))
                wires.extend(shapeWires)
            self.wires = wires
            Path.Log.debug("processing {} jobshapes -> {} wires".format(len(jobshapes), len(wires)))
        # the last command is a move to clearance, which is automatically added by PathOp
        if self.commandlist:
            self.commandlist.pop()

    def opUpdateDepths(self, obj):
        """updateDepths(obj) ... engraving is always done at the top most z-value"""
        job = PathUtils.findParentJob(obj)
        self.opSetDefaultValues(obj, job)


def SetupProperties():
    return ["StartVertex"]


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns an Engrave operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    obj.Proxy = ObjectEngrave(obj, name, parentJob)
    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM Truss command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Truss:
    "the Arch Truss command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Truss",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Truss", "Truss"),
            "Accel": "T, U",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Truss", "Creates a truss object from the selected line or from scratch"
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        self.doc = FreeCAD.ActiveDocument
        sel = FreeCADGui.Selection.getSelection()
        if len(sel) > 1:
            FreeCAD.Console.PrintError(
                translate("Arch", "Select only one base object or none") + "\n"
            )
        elif len(sel) == 1:
            # build on selection
            basename = "FreeCAD.ActiveDocument." + FreeCADGui.Selection.getSelection()[0].Name
            self.createTruss(basename)
        else:
            # interactive line drawing
            import WorkingPlane

            FreeCAD.activeDraftCommand = self  # register as a Draft command for auto grid on/off
            self.points = []
            WorkingPlane.get_working_plane()
            if hasattr(FreeCADGui, "Snapper"):
                FreeCADGui.Snapper.getPoint(callback=self.getPoint)

    def getPoint(self, point=None, obj=None):
        """Callback for clicks during interactive mode"""

        if point is None:
            # cancelled
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            return
        self.points.append(point)
        if len(self.points) == 1:
            FreeCADGui.Snapper.getPoint(last=self.points[0], callback=self.getPoint)
        elif len(self.points) == 2:
            FreeCAD.activeDraftCommand = None
            FreeCADGui.Snapper.off()
            self.createTruss()

    def createTruss(self, basename=""):
        """Creates the truss"""

        FreeCADGui.Control.closeDialog()
        self.doc.openTransaction(translate("Arch", "Create Truss"))
        FreeCADGui.addModule("Draft")
        FreeCADGui.addModule("Arch")
        if not basename:
            if self.points:
                cmd = "base = Draft.makeLine(FreeCAD."
                cmd += str(self.points[0]) + ",FreeCAD." + str(self.points[1]) + ")"
                FreeCADGui.doCommand(cmd)
                basename = "base"
        FreeCADGui.doCommand("obj = Arch.makeTruss(" + basename + ")")
        FreeCADGui.doCommand("Draft.autogroup(obj)")
        self.doc.commitTransaction()
        self.doc.recompute()


FreeCADGui.addCommand("Arch_Truss", Arch_Truss())



# **************************************************************************
#   Copyright (c) 2024 Syres                                               *
#                                                                          *
#   This file is part of the FreeCAD CAx development system.               *
#                                                                          *
#   FreeCAD is free software: you can redistribute it and/or modify it     *
#   under the terms of the GNU Lesser General Public License as            *
#   published by the Free Software Foundation, either version 2.1 of the   *
#   License, or (at your option) any later version.                        *
#                                                                          *
#   FreeCAD is distributed in the hope that it will be useful, but         *
#   WITHOUT ANY WARRANTY; without even the implied warranty of             *
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#   Lesser General Public License for more details.                        *
#                                                                          *
#   You should have received a copy of the GNU Lesser General Public       *
#   License along with FreeCAD. If not, see                                *
#   <https://www.gnu.org/licenses/>.                                       *
# **************************************************************************

import unittest
import FreeCAD
import Part
import Sketcher
from FreeCAD import Vector

App = FreeCAD


class TestSketchValidateCoincidents(unittest.TestCase):
    """Sketch validation test"""

    def setUp(self):
        self.Doc = FreeCAD.newDocument("SketchValidateCoincidentsTest")

    def testSingleMissingCoincidentCase(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        v0 = Vector(-47.680691, 18.824165000000004, 0.0)
        v1 = Vector(-47.680691, -27.346279, 0.0)
        v2 = Vector(51.132679, -27.346279, 0.0)
        v3 = Vector(51.132679, 18.824165000000004, 0.0)
        v4 = Vector(51.132679, 18.824165, 0.0)
        v5 = Vector(-47.680691, 18.824165, 0.0)

        geo0 = sketch.addGeometry(Part.LineSegment(v0, v1))
        geo1 = sketch.addGeometry(Part.LineSegment(v1, v2))
        geo2 = sketch.addGeometry(Part.LineSegment(v2, v3))
        geo3 = sketch.addGeometry(Part.LineSegment(v4, v5))

        sketch.addConstraint(Sketcher.Constraint("Coincident", geo0, 2, geo1, 1))
        sketch.addConstraint(Sketcher.Constraint("Coincident", geo1, 2, geo2, 1))
        sketch.addConstraint(Sketcher.Constraint("Vertical", geo0))
        sketch.addConstraint(Sketcher.Constraint("Vertical", geo2))
        sketch.addConstraint(Sketcher.Constraint("Horizontal", geo1))
        sketch.addConstraint(Sketcher.Constraint("Coincident", geo3, 1, geo2, 2))
        sketch.addConstraint(Sketcher.Constraint("Horizontal", geo3))
        self.Doc.recompute()
        self.assertEqual(sketch.ConstraintCount, 7)
        sketch.detectMissingPointOnPointConstraints()
        sketch.makeMissingPointOnPointCoincident()
        self.Doc.recompute()
        self.assertEqual(sketch.ConstraintCount, 8)
        del v0, v1, v2, v3, v4, v5
        del geo0, geo1, geo2, geo3
        del sketch

    def testDegenratedGeometryCase(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        v0 = Vector(-47.680691, 18.824165000000004, 0.0)
        v1 = Vector(-47.680691, -27.346279, 0.0)
        v2 = Vector(-47.680691, -27.34627900001, 0.0)

        geo0 = sketch.addGeometry(Part.LineSegment(v0, v1))
        geo1 = sketch.addGeometry(Part.LineSegment(v1, v2))
        sketch.addConstraint(Sketcher.Constraint("Coincident", geo0, 2, geo1, 1))
        self.Doc.recompute()
        tol = 1.0e-8
        self.assertEqual(sketch.ConstraintCount, 1)
        self.assertEqual(sketch.detectDegeneratedGeometries(tol), 1)
        self.assertEqual(sketch.removeDegeneratedGeometries(tol), 1)
        self.assertEqual(sketch.detectDegeneratedGeometries(tol), 0)
        self.assertEqual(sketch.ConstraintCount, 0)
        del v0, v1, v2
        del geo0, geo1
        del sketch

    def testDeleteConstraintsToExternalCase(self):
        box = self.Doc.addObject("Part::Box", "Box")
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        self.Doc.recompute()
        sketch.addExternal("Box", "Edge12")

        v0 = Vector(-47.680691, 18.824165000000004, 0.0)
        v1 = Vector(-47.680691, -27.346279, 0.0)
        geo0 = sketch.addGeometry(Part.LineSegment(v0, v1))
        sketch.addConstraint(Sketcher.Constraint("Horizontal", geo0))
        sketch.addConstraint(Sketcher.Constraint("Equal", -3, geo0))
        self.Doc.recompute()
        self.assertEqual(sketch.ConstraintCount, 2)
        sketch.delConstraintsToExternal()
        self.Doc.recompute()
        self.assertEqual(sketch.ConstraintCount, 1)
        del v0, v1
        del geo0
        del sketch
        del box

    def testValidateConstraintsCase(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        v0 = Vector(-47.680691, 18.824165000000004, 0.0)
        v1 = Vector(-47.680691, -27.346279, 0.0)
        v2 = Vector(51.132679, -27.346279, 0.0)
        geo0 = sketch.addGeometry(Part.LineSegment(v0, v1))
        geo1 = sketch.addGeometry(Part.LineSegment(v1, v2))
        sketch.addConstraint(Sketcher.Constraint("Coincident", geo0, 2, geo1, 1))

        c = sketch.Constraints[0]
        c.First = 2
        sketch.Constraints = [c]
        self.assertEqual(sketch.ConstraintCount, 1)
        self.assertFalse(sketch.evaluateConstraints())

        sketch.validateConstraints()

        self.assertEqual(sketch.ConstraintCount, 0)
        self.assertTrue(sketch.evaluateConstraints())
        del v0, v1, v2
        del geo0, geo1
        del sketch, c

    def testExternalGeoDeletion(self):
        """Make sure that we don't remove External Geometry references to deleted geometry.
        See https://github.com/FreeCAD/FreeCAD/issues/16361"""
        if "BUILD_PARTDESIGN" in FreeCAD.__cmake__:
            doc = App.ActiveDocument
            doc.addObject("PartDesign::Body", "Body")
            doc.Body.Label = "Body"
            doc.recompute()
            doc.Body.newObject("Sketcher::SketchObject", "Sketch")
            doc.Sketch.AttachmentSupport = (doc.XY_Plane, [""])
            doc.Sketch.MapMode = "FlatFace"
            doc.recompute()
            geoList = []
            geoList.append(
                Part.LineSegment(
                    App.Vector(7.548310, 5.193216, 0.000000),
                    App.Vector(31.461353, 5.193216, 0.000000),
                )
            )
            geoList.append(
                Part.LineSegment(
                    App.Vector(31.461353, 5.193216, 0.000000),
                    App.Vector(31.461353, 20.652151, 0.000000),
                )
            )
            geoList.append(
                Part.LineSegment(
                    App.Vector(31.461353, 20.652151, 0.000000),
                    App.Vector(7.548310, 20.652151, 0.000000),
                )
            )
            geoList.append(
                Part.LineSegment(
                    App.Vector(7.548310, 20.652151, 0.000000),
                    App.Vector(7.548310, 5.193216, 0.000000),
                )
            )
            doc.Sketch.addGeometry(geoList, False)
            del geoList
            constraintList = []
            constraintList.append(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
            constraintList.append(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
            constraintList.append(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
            constraintList.append(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
            constraintList.append(Sketcher.Constraint("Horizontal", 0))
            constraintList.append(Sketcher.Constraint("Horizontal", 2))
            constraintList.append(Sketcher.Constraint("Vertical", 1))
            constraintList.append(Sketcher.Constraint("Vertical", 3))
            doc.Sketch.addConstraint(constraintList)
            del constraintList
            constraintList = []
            doc.recompute()
            doc.Body.newObject("Sketcher::SketchObject", "Sketch001")
            doc.Sketch001.AttachmentSupport = (doc.XY_Plane, [""])
            doc.Sketch001.MapMode = "FlatFace"
            doc.recompute()
            doc.Sketch001.addExternal("Sketch", "Edge3")
            geoList = []
            geoList.append(
                Part.LineSegment(
                    App.Vector(33.192780, 22.530144, 0.000000),
                    App.Vector(38.493977, 32.289181, 0.000000),
                )
            )
            doc.Sketch001.addGeometry(geoList, False)
            del geoList
            constraintList = []
            doc.Sketch001.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -3, 1))
            doc.recompute()
            doc.Sketch.delGeometries([1])
            doc.Sketch.delGeometries([1])
            doc.recompute()
            doc.Sketch001.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -3, 1))
            doc.recompute()
            # Assert
            self.assertEqual(len(doc.Sketch001.Constraints), 2)  # Still have the constraints
            self.assertEqual(len(doc.Sketch001.ExternalGeometry), 0)
            self.assertEqual(len(doc.Sketch001.Geometry), 1)
            self.assertEqual(
                len(doc.Sketch001.ExternalGeo), 3
            )  # Two axis, plus one the reference to deleted geometry

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument(self.Doc.Name)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2018 Benjamin Alterauge (ageeye)                        *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create PointArray objects.

The copies will be created at the points of a point object.
"""
## @package gui_pointarray
# \ingroup draftguitools
# \brief Provides GUI tools to create PointArray objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import Draft_rc
import draftguitools.gui_base_original as gui_base_original

from draftutils.messages import _err
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class PointArray(gui_base_original.Modifier):
    """Gui Command for the Point array tool.

    Parameters
    ----------
    use_link: bool, optional
        It defaults to `False`. If it is `True`, the created object
        will be a `Point link array`.
    """

    def __init__(self, use_link=False):
        super(PointArray, self).__init__()
        self.use_link = use_link

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_PointArray",
            "MenuText": QT_TRANSLATE_NOOP("Draft_PointArray", "Point Array"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_PointArray",
                "Creates copies of the selected object at the points of a point object",
            ),
        }

    def Activated(self, name="Point array"):
        """Execute when the command is called."""
        super(PointArray, self).Activated(name=name)
        # This was deactivated because it doesn't work correctly;
        # the selection needs to be made on two objects, but currently
        # it only selects one.

        # if not Gui.Selection.getSelectionEx():
        #     if self.ui:
        #         self.ui.selectUi()
        #         _msg(translate("draft",
        #                        "Please select exactly two objects, "
        #                        "the base object and the point object, "
        #                        "before calling this command."))
        #         self.call = \
        #             self.view.addEventCallback("SoEvent",
        #                                        gui_tool_utils.selectObject)
        # else:
        #     self.proceed()
        self.proceed()

    def proceed(self):
        """Proceed with the command if one object was selected."""
        sel = Gui.Selection.getSelectionEx()
        if len(sel) != 2:
            _err(
                translate(
                    "draft",
                    "Select exactly 2 objects, the base object and the point object, before calling this command",
                )
            )
        else:
            base_object = sel[0].Object
            point_object = sel[1].Object
            extra = None

            Gui.addModule("Draft")
            _cmd = "Draft.make_point_array"
            _cmd += "("
            _cmd += "App.ActiveDocument." + base_object.Name + ", "
            _cmd += "App.ActiveDocument." + point_object.Name + ", "
            _cmd += "extra=" + str(extra) + ", "
            _cmd += "use_link=" + str(self.use_link)
            _cmd += ")"

            _cmd_list = [
                "_obj_ = " + _cmd,
                "Draft.autogroup(_obj_)",
                "App.ActiveDocument.recompute()",
            ]
            self.commit(translate("draft", "Create Point Array"), _cmd_list)

        # Commit the transaction and execute the commands
        # through the parent class
        self.finish()


Gui.addCommand("Draft_PointArray", PointArray())


class PointLinkArray(PointArray):
    """Gui Command for the PointLinkArray tool based on the PointArray tool."""

    def __init__(self):
        super(PointLinkArray, self).__init__(use_link=True)

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_PointLinkArray",
            "MenuText": QT_TRANSLATE_NOOP("Draft_PointLinkArray", "Point Link Array"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_PointLinkArray",
                "Creates linked copies of the selected object at the points of a point object",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super(PointLinkArray, self).Activated(name="Point link array")


Gui.addCommand("Draft_PointLinkArray", PointLinkArray())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009 Yorik van Havre <yorik@uncreated.net>                        *
# *   (c) 2010 Ken Cline <cline@frii.com>                                   *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create Rectangle objects."""
## @package gui_rectangles
# \ingroup draftguitools
# \brief Provides GUI tools to create Rectangle objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import DraftVecUtils
from draftguitools import gui_base_original
from draftguitools import gui_tool_utils
from draftguitools import gui_trackers as trackers
from draftutils import params
from draftutils import utils
from draftutils.messages import _err, _toolmsg
from draftutils.translate import translate


class Rectangle(gui_base_original.Creator):
    """Gui command for the Rectangle tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Rectangle",
            "Accel": "R, E",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Rectangle", "Rectangle"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Rectangle", "Creates a 2-point rectangle"),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Rectangle")
        if self.ui:
            self.refpoint = None
            self.ui.pointUi(title=translate("draft", "Rectangle"), icon="Draft_Rectangle")
            self.ui.extUi()
            self.call = self.view.addEventCallback("SoEvent", self.action)
            self.rect = trackers.rectangleTracker()
            _toolmsg(translate("draft", "Pick first point"))

    def finish(self, cont=False):
        """Terminate the operation.

        Parameters
        ----------
        cont: bool or None, optional
            Restart (continue) the command if `True`, or if `None` and
            `ui.continueMode` is `True`.
        """
        self.end_callbacks(self.call)
        if self.ui:
            self.rect.off()
            self.rect.finalize()
        super().finish()
        if cont or (cont is None and self.ui and self.ui.continueMode):
            self.Activated()

    def createObject(self):
        """Create the final object in the current document."""
        p1 = self.node[0]
        p3 = self.node[-1]
        diagonal = p3.sub(p1)
        p2 = p1.add(DraftVecUtils.project(diagonal, self.wp.v))
        p4 = p1.add(DraftVecUtils.project(diagonal, self.wp.u))
        length = p4.sub(p1).Length
        if abs(DraftVecUtils.angle(p4.sub(p1), self.wp.u, self.wp.axis)) > 1:
            length = -length
        height = p2.sub(p1).Length
        if abs(DraftVecUtils.angle(p2.sub(p1), self.wp.v, self.wp.axis)) > 1:
            height = -height
        try:
            # The command to run is built as a series of text strings
            # to be committed through the `draftutils.todo.ToDo` class.
            rot, sup, pts, fil = self.getStrings()
            base = p1
            if length < 0:
                length = -length
                base = base.add((p1.sub(p4)).negative())
            if height < 0:
                height = -height
                base = base.add((p1.sub(p2)).negative())
            Gui.addModule("Draft")
            if params.get_param("UsePartPrimitives"):
                # Insert a Part::Primitive object
                _cmd = "FreeCAD.ActiveDocument."
                _cmd += 'addObject("Part::Plane", "Plane")'
                _cmd_list = [
                    "plane = " + _cmd,
                    "plane.Length = " + str(length),
                    "plane.Width = " + str(height),
                    "pl = FreeCAD.Placement()",
                    "pl.Rotation.Q=" + rot,
                    "pl.Base = " + DraftVecUtils.toString(base),
                    "plane.Placement = pl",
                    "Draft.autogroup(plane)",
                    "Draft.select(plane)",
                    "FreeCAD.ActiveDocument.recompute()",
                ]
                self.commit(translate("draft", "Create Plane"), _cmd_list)
            else:
                _cmd = "Draft.make_rectangle"
                _cmd += "("
                _cmd += "length=" + str(length) + ", "
                _cmd += "height=" + str(height) + ", "
                _cmd += "placement=pl, "
                _cmd += "face=" + fil + ", "
                _cmd += "support=" + sup
                _cmd += ")"
                _cmd_list = [
                    "pl = FreeCAD.Placement()",
                    "pl.Rotation.Q = " + rot,
                    "pl.Base = " + DraftVecUtils.toString(base),
                    "rec = " + _cmd,
                    "Draft.autogroup(rec)",
                    "FreeCAD.ActiveDocument.recompute()",
                ]
                self.commit(translate("draft", "Create Rectangle"), _cmd_list)
        except Exception:
            _err("Draft: error delaying commit")
        self.finish(cont=None)

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif not self.ui.mouse:
            pass
        elif arg["Type"] == "SoLocation2Event":  # mouse movement detection
            self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg, noTracker=True)
            self.rect.update(self.point)
            gui_tool_utils.redraw3DView()
        elif (
            arg["Type"] == "SoMouseButtonEvent"
            and arg["State"] == "DOWN"
            and arg["Button"] == "BUTTON1"
        ):

            if arg["Position"] == self.pos:
                self.finish(cont=None)
                return

            if (not self.node) and (not self.support):
                gui_tool_utils.getSupport(arg)
                self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg, noTracker=True)
            if self.point:
                self.ui.redraw()
                self.pos = arg["Position"]
                self.appendPoint(self.point)

    def numericInput(self, numx, numy, numz):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.point = App.Vector(numx, numy, numz)
        self.appendPoint(self.point)

    def appendPoint(self, point):
        """Append a point to the list of nodes."""
        self.node.append(point)
        if len(self.node) > 1:
            self.rect.update(point)
            self.createObject()
        else:
            _toolmsg(translate("draft", "Pick opposite point"))
            self.ui.setRelative()
            self.rect.setorigin(point)
            self.rect.on()
            if self.planetrack:
                self.planetrack.set(point)
            self.update_hints()

    def get_hints(self):
        if len(self.node) == 0:
            hints = [
                Gui.InputHint(translate("draft", "%1 pick first point"), Gui.UserInput.MouseLeft)
            ]
        else:
            hints = [
                Gui.InputHint(translate("draft", "%1 pick opposite point"), Gui.UserInput.MouseLeft)
            ]
        return (
            hints
            + gui_tool_utils._get_hint_xyz_constrain()
            + gui_tool_utils._get_hint_mod_constrain()
            + gui_tool_utils._get_hint_mod_snap()
        )


Gui.addCommand("Draft_Rectangle", Rectangle())

## @}



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import sys
import FreeCAD
from FreeCAD import Vector

import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Flux - Elmer",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["electrostatic potential", "temperature"],
        "solvers": ["elmer"],
        "material": "solid",
        "equations": ["electrostatic", "flux", "heat"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.equation_flux_elmer import setup
setup()

Potential flux and heat flux - Elmer solver

"""
    )


def setup(doc=None, solvertype="elmer"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric object
    cube = doc.addObject("Part::Box", "Cube")

    # set view
    doc.recompute()
    if FreeCAD.GuiUp:
        cube.ViewObject.Transparency = 50
        cube.ViewObject.Document.activeView().viewAxonometric()
        cube.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")
    if FreeCAD.GuiUp:
        import FemGui

        FemGui.setActiveAnalysis(analysis)

    # solver
    if solvertype == "elmer":
        solver_obj = ObjectsFem.makeSolverElmer(doc, "SolverElmer")
        ObjectsFem.makeEquationElectrostatic(doc, solver_obj)
        ObjectsFem.makeEquationHeat(doc, solver_obj)
        equation_flux_potential = ObjectsFem.makeEquationFlux(doc, solver_obj)
        equation_flux_temperature = ObjectsFem.makeEquationFlux(doc, solver_obj)
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
        return doc
    analysis.addObject(solver_obj)

    # solver settings
    equation_flux_potential.Label = "Flux_Potential"
    equation_flux_potential.FluxCoefficient = "None"
    equation_flux_potential.FluxVariable = "Potential"
    equation_flux_temperature.Label = "Flux_Temperature"

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Graphite")
    mat = material_obj.Material
    mat["Name"] = "Graphite"
    mat["Density"] = "1750.0 kg/m^3"
    mat["PoissonRatio"] = "0.20"
    mat["ThermalConductivity"] = "96 W/m/K"
    mat["ThermalExpansionCoefficient"] = "8 µm/m/K"
    mat["SpecificHeat"] = "720.00 J/kg/K"
    mat["RelativePermittivity"] = "0.7"
    mat["ElectricalConductivity"] = "2e4 S/m"
    mat["RelativePermeability"] = "1.0"
    material_obj.Material = mat
    material_obj.References = [(cube, "Solid1")]
    analysis.addObject(material_obj)

    # constraint potential 1V
    Potential_1V = ObjectsFem.makeConstraintElectrostaticPotential(doc, "Potential_1V")
    Potential_1V.Potential = "1 V"
    Potential_1V.NormalDirection = Vector(0, -1, 0)
    Potential_1V.References = [(cube, "Face3")]
    analysis.addObject(Potential_1V)

    # constraint potential 0V
    Potential_0V = ObjectsFem.makeConstraintElectrostaticPotential(doc, "Potential_0V")
    Potential_0V.Potential = "0 V"
    Potential_0V.NormalDirection = Vector(0, 0, 0)
    Potential_0V.References = [(cube, "Face4")]
    analysis.addObject(Potential_0V)

    # constraint wall temperature
    Temperature_300K = ObjectsFem.makeConstraintTemperature(doc, "Temperature_300K")
    Temperature_300K.Temperature = 300.0
    Temperature_300K.NormalDirection = Vector(-1, 0, 0)
    Temperature_300K.References = [(cube, "Face1")]
    analysis.addObject(Temperature_300K)

    # constraint inlet temperature
    Temperature_600K = ObjectsFem.makeConstraintTemperature(doc, "Temperature_600K")
    Temperature_600K.Temperature = 600.0
    Temperature_600K.NormalDirection = Vector(1, 0, 0)
    Temperature_600K.References = [(cube, "Face2")]
    analysis.addObject(Temperature_600K)

    # mesh
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.Shape = cube
    femmesh_obj.CharacteristicLengthMax = "1 mm"
    femmesh_obj.ViewObject.Visibility = False

    # generate the mesh
    generate_mesh.mesh_from_mesher(femmesh_obj, "gmsh")

    doc.recompute()
    return doc



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint pressure"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import FreeCAD


def get_analysis_types():
    return ["buckling", "static", "thermomech"]


def get_sets_name():
    return "constraints_pressure_element_face_loads"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def write_meshdata_constraint(f, femobj, prs_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    if prs_obj.EnableAmplitude:
        f.write(f"*DLOAD, AMPLITUDE={prs_obj.Name}\n")
    else:
        f.write("*DLOAD\n")
    rev = -1 if prs_obj.Reversed else 1
    # the pressure has to be output in MPa
    pressure = prs_obj.Pressure.getValueAs("MPa").Value
    pressure *= rev
    for feat, surf, is_sub_el in femobj["PressureFaces"]:
        f.write("** {0.Name}.{1[0]}\n".format(*feat))
        if is_sub_el:
            for elem, fno in surf:
                f.write(f"{elem},P{fno},{pressure}\n")
        else:
            for elem in surf:
                f.write(f"{elem},P,{-1*pressure}\n")
